import os
import sys
import sqlite3
from math import prod, log2
from numpy import argsort
import waLBerla as wlb
from waLBerla.tools.config import get_best_3d_decomposition
from waLBerla.tools.sqlitedb import sequenceValuesToScalars, checkAndUpdateSchema, storeSingle

try:
    import machinestate as ms
except ImportError:
    ms = None

# Define global variables with default values
BENCHMARK = int(os.environ.get('BENCHMARK', 0))
CLUSTER = str(os.environ.get('CLUSTER', None))
DB_FILE = str(os.environ.get('DB_FILE', "cpu_benchmark.sqlite3"))
MINIMAL_RUNTIME_SINGLE_NODE = int(os.environ.get('MINIMAL_RUNTIME_SINGLE_NODE', 5))

TIME_STEPS_FOR_128_BLOCK = None
TIME_STEPS_TO_EXECUTE = None
PROCESSES_PER_NODE = None
MAIN_MEMORY_GB, NUM_CORES_PER_NODE, MAX_AVAIL_NODES = None, None, None
WeakX, WeakY, WeakZ = None, None, None
StrongDecompositionX, StrongDecompositionY, StrongDecompositionZ = None, None, None
DOMAIN_BLOCK_SIZES_SINGLE_NODE = None


def set_parameter_for_specific_cluster():
    global TIME_STEPS_FOR_128_BLOCK, TIME_STEPS_TO_EXECUTE, PROCESSES_PER_NODE
    global MAIN_MEMORY_GB, NUM_CORES_PER_NODE, MAX_AVAIL_NODES
    global WeakX, WeakY, WeakZ
    global StrongDecompositionX, StrongDecompositionY, StrongDecompositionZ
    global DOMAIN_BLOCK_SIZES_SINGLE_NODE

    cluster_mapping = {
        "marenostrum5": "MareNostrum5",
        "marenostrum": "MareNostrum5",
        "lumi": "LUMI"
    }

    normalized_cluster = (CLUSTER or "MareNostrum5").lower() # MareNostrum5 is the default choice if CLUSTER=None
    matched_cluster = cluster_mapping.get(normalized_cluster)

    if matched_cluster == "MareNostrum5":
        TIME_STEPS_FOR_128_BLOCK = int(os.environ.get('TIME_STEPS_FOR_128_BLOCK', 10))
        PROCESSES_PER_NODE = int(os.environ.get('PROCESSES_PER_NODE', 112))
        TIME_STEPS_TO_EXECUTE = int(os.environ.get('TIME_STEPS_TO_EXECUTE', 5))
        MAIN_MEMORY_GB = int(os.environ.get('MAIN_MEMORY_GB', 256)) # Default for MareNostrum5-GPP
        NUM_CORES_PER_NODE = int(os.environ.get('NUM_CORES_PER_NODE', 112)) # Default for MareNostrum5-GPP
        MAX_AVAIL_NODES = int(os.environ.get('MAX_AVAIL_NODES', 100)) # Default for MareNostrum5-GPP

        WeakX = int(os.environ.get('WeakX', 224))
        WeakY = int(os.environ.get('WeakY', 224))
        WeakZ = int(os.environ.get('WeakZ', 112))

        StrongDecompositionX = int(os.environ.get('StrongX', 4))
        StrongDecompositionY = int(os.environ.get('StrongY', 4))
        StrongDecompositionZ = int(os.environ.get('StrongZ', 7))

        DOMAIN_BLOCK_SIZES_SINGLE_NODE = [
            (224, 208, 160),    # N=8.349e+08 ~ 98.6% | 28.6%
            (224, 176, 176),    # N=7.771e+08 ~ 91.8% | 21.4%
            (240, 240, 112),    # N=7.225e+08 ~ 85.3% | 57.1% # 100 nodes included
            (224, 160, 160),    # N=6.423e+08 ~ 75.9% | 28.6% # 100 nodes included
            (240, 160, 112),    # N=4.817e+08 ~ 56.9% | 57.1% # 100 nodes included
            (128, 128, 112),    # N=2.055e+08 ~ 24.3% |  7.1%
            (112, 112, 112),    # N=1.574e+08 ~ 18.6% |  1.0%
            (112, 80, 80),      # N=8.028e+07 ~  9.5% | 14.3%
        ]
        return True

    elif matched_cluster == "LUMI":
        TIME_STEPS_FOR_128_BLOCK = int(os.environ.get('TIME_STEPS_FOR_128_BLOCK', 2))
        PROCESSES_PER_NODE = int(os.environ.get('PROCESSES_PER_NODE', 128))
        TIME_STEPS_TO_EXECUTE = int(os.environ.get('TIME_STEPS_TO_EXECUTE', 5))
        MAIN_MEMORY_GB = int(os.environ.get('MAIN_MEMORY_GB', 256)) # Default for MareNostrum5-GPP
        NUM_CORES_PER_NODE = int(os.environ.get('NUM_CORES_PER_NODE', 128)) # Default for MareNostrum5-GPP
        MAX_AVAIL_NODES = int(os.environ.get('MAX_AVAIL_NODES', 512)) # Default for MareNostrum5-GPP

        WeakX = int(os.environ.get('WeakX', 192))
        WeakY = int(os.environ.get('WeakY', 192))
        WeakZ = int(os.environ.get('WeakZ', 160))

        StrongDecompositionX = int(os.environ.get('StrongX', 4))
        StrongDecompositionY = int(os.environ.get('StrongY', 4))
        StrongDecompositionZ = int(os.environ.get('StrongZ', 8))

        DOMAIN_BLOCK_SIZES_SINGLE_NODE = [
            (192, 192, 176),    # N=8.305e+08 ~ 98.1% |  8.3%
            (192, 192, 160),    # N=7.550e+08 ~ 89.2% |  ?.?%
            (192, 160, 160),    # N=6.291e+08 ~ 74.3% | 16.7%
            (160, 160, 128),    # N=4.194e+08 ~ 49.5% | 16.7%
            (128, 128, 128),    # N=2.684e+08 ~ 31.7% |  1.0%
            (128, 96, 96),      # N=1.510e+08 ~ 17.8% | 16.7%
            (64, 64, 64),       # N=3.355e+07 ~  4.0% |  1.0%
        ]
        return True

    else:
        wlb.log_warning_on_root(f"Cluster '{str(CLUSTER)}' is not recognized. Benchmark execution skipped.")
        return False



def num_time_steps(block_size, time_steps_for_128_block=None):
    """
    Calculate the number of time steps based on the block size.

    This function computes the number of time steps required for a given block size
    by scaling the time steps that could be executed on one process within one second
    for a 128x128x128 cells_per_block to the given cells_per_block size.

    Parameters:
    block_size (tuple): A tuple of three integers representing the dimensions of the cells_per_block (x, y, z).
    time_steps_for_128_block (int, optional): The number of time steps for a 128x128x128 block.

    Returns:
    int: The calculated number of time steps, with a minimum value of 5.
    """
    if time_steps_for_128_block is None:
        time_steps_for_128_block = TIME_STEPS_FOR_128_BLOCK
    if time_steps_for_128_block is None:
        raise ValueError("TIME_STEPS_FOR_128_BLOCK is not initialized.")

    cells = block_size[0] * block_size[1] * block_size[2]
    time_steps = (128 ** 3 / cells) * time_steps_for_128_block * MINIMAL_RUNTIME_SINGLE_NODE
    if time_steps < 5:
        time_steps = 5
    return int(time_steps)


ldc_setup = {'Border': [
    {'direction': 'N', 'walldistance': -1, 'flag': 'UBB'},
    {'direction': 'W', 'walldistance': -1, 'flag': 'NoSlip'},
    {'direction': 'E', 'walldistance': -1, 'flag': 'NoSlip'},
    {'direction': 'S', 'walldistance': -1, 'flag': 'NoSlip'},
    {'direction': 'B', 'walldistance': -1, 'flag': 'NoSlip'},
    {'direction': 'T', 'walldistance': -1, 'flag': 'NoSlip'},
]}


class Scenario:
    def __init__(self, benchmark_name="not defined", blocks=None, cells_per_block=(128, 128, 128), periodic=(1, 1, 1), blocks_per_process=1,
                 timesteps=None, time_step_strategy="normal", omega=1.8, inner_outer_split=(1, 1, 1),
                 warmup_steps=2, outer_iterations=3, init_shear_flow=False, boundary_setup=False,
                 vtk_write_frequency=0, remaining_time_logger_frequency=-1, db_file_name=None):

        self.benchmark_name = benchmark_name

        if boundary_setup:
            init_shear_flow = False
            periodic = (0, 0, 0)

        self.blocks_per_process = blocks_per_process
        if blocks is None:
            self.blocks = tuple(get_best_3d_decomposition(self.blocks_per_process * wlb.mpi.numProcesses()))
        else:
            self.blocks = tuple(blocks)

        self.cells_per_block = tuple(cells_per_block)
        self.periodic = tuple(periodic)

        self.time_step_strategy = time_step_strategy
        self.omega = omega
        self.timesteps = timesteps if timesteps else num_time_steps(cells_per_block)
        self.inner_outer_split = tuple(inner_outer_split)
        self.init_shear_flow = init_shear_flow
        self.boundary_setup = boundary_setup
        self.warmup_steps = warmup_steps
        self.outer_iterations = outer_iterations

        self.vtk_write_frequency = vtk_write_frequency
        self.remaining_time_logger_frequency = remaining_time_logger_frequency
        self.db_file_name = DB_FILE if db_file_name is None else db_file_name

        self.config_dict = self.config(print_dict=False)

    @wlb.member_callback
    def config(self, print_dict=True):
        from pprint import pformat
        config_dict = {
            'DomainSetup': {
                'blocks': self.blocks,
                'cellsPerBlock': self.cells_per_block,
                'periodic': self.periodic,
                'cartesianSetup': (self.blocks_per_process == 1)
            },
            'Parameters': {
                'omega': self.omega,
                'warmupSteps': self.warmup_steps,
                'outerIterations': self.outer_iterations,
                'timeStepStrategy': self.time_step_strategy,
                'timesteps': self.timesteps,
                'initShearFlow': self.init_shear_flow,
                'innerOuterSplit': self.inner_outer_split,
                'vtkWriteFrequency': self.vtk_write_frequency,
                'remainingTimeLoggerFrequency': self.remaining_time_logger_frequency
            }
        }
        if self.boundary_setup:
            config_dict["Boundaries"] = ldc_setup

        if print_dict:
            wlb.log_info_on_root("\n" + 50*"=" + "\n")
            wlb.log_info_on_root("Scenario:\n" + pformat(config_dict))
        return config_dict

    @wlb.member_callback
    def results_callback(self, **kwargs):
        data = {}
        data.update(self.config_dict['Parameters'])
        data.update(self.config_dict['DomainSetup'])
        data.update(kwargs)

        data['Identifier'] = self.benchmark_name
        data['executable'] = sys.argv[0]
        data['compile_flags'] = wlb.build_info.compiler_flags
        data['walberla_version'] = wlb.build_info.version
        data['build_machine'] = wlb.build_info.build_machine

        if ms:
            state = ms.MachineState(extended=False, anonymous=True)
            state.generate()  # generate subclasses
            state.update()  # read information
            data["MachineState"] = str(state.get())
        else:
            print("MachineState module is not available. MachineState was not saved")

        sequenceValuesToScalars(data)

        result = data
        sequenceValuesToScalars(result)
        num_tries = 4
        table_name = "timeStepping_" + str(self.time_step_strategy)
        # table_name = f"CPU_runs_{data['stencil']}_{data['streamingPattern']}_{data['collisionSetup']}_{prod(self.blocks)}nodes_{'x'.join(map(str, self.cells_per_block))}cellsPerBlock"
        table_name = table_name.replace("-", "_")
        table_name = table_name.replace(" ", "_")

        for num_try in range(num_tries):
            try:
                checkAndUpdateSchema(result, table_name, self.db_file_name, alter_table=True)
                storeSingle(result, table_name, self.db_file_name)
                break
            except sqlite3.OperationalError as e:
                wlb.log_warning(f"Sqlite DB writing failed: try {num_try + 1}/{num_tries}  {str(e)}")


# -------------------------------------- Functions trying different parameter sets -----------------------------------


def weak_scaling_benchmark():
    wlb.log_info_on_root("Running weak scaling benchmark with one block per proc and communication hiding")
    wlb.log_info_on_root("")

    scenarios = wlb.ScenarioManager()
    cells_per_block = (WeakX, WeakY, WeakZ)

    for t in ["simpleOverlap"]:
        for split in [(1, 1, 1), (8, 8, 8), (16, 1, 1)]:
            if all(2 * s < c for s, c in zip(split, cells_per_block)):
                scenarios.add(Scenario(benchmark_name="weakScalingUniformGrid",
                                       cells_per_block=tuple(cells_per_block),
                                       time_step_strategy=t,
                                       inner_outer_split=tuple(split),
                                       boundary_setup=True,
                                       outer_iterations=2,
                                       timesteps=TIME_STEPS_TO_EXECUTE,
                ))
            else:
                wlb.log_warning_on_root(f"Split {split} would exceed cells per block {cells_per_block}. Skipping.")


def strong_scaling_benchmark():
    wlb.log_info_on_root("Running strong scaling benchmark with one block per proc with communication hiding")
    wlb.log_info_on_root("")

    scenarios = wlb.ScenarioManager()

    num_processes = wlb.mpi.numProcesses()
    initial_blocks = [StrongDecompositionX, StrongDecompositionY, StrongDecompositionZ]
    processes_per_node = prod(initial_blocks)
    assert num_processes % processes_per_node == 0, "Number of processes must be a multiple of processes per node."
    number_of_nodes = num_processes // processes_per_node
    initial_cells_per_block = [WeakX, WeakY, WeakZ]
    scaling_step = log2(number_of_nodes)

    if not scaling_step.is_integer():
        blocks = None
        total_domain_size = prod(initial_cells_per_block) * processes_per_node
        cells_per_block = tuple(get_best_3d_decomposition(total_domain_size, descending=True))
    else: # This scaling step has a power of two number of nodes.
        blocks = initial_blocks
        cells_per_block = initial_cells_per_block
        for _ in range(int(scaling_step)):
            division_indices = argsort(cells_per_block)[::-1] # Sort in descending order
            for idx in division_indices:
                if cells_per_block[idx] % 2 == 0:
                    cells_per_block[idx] //= 2
                    blocks[idx] *= 2
                    break
            else: # else block runs only if the loop completes without encountering a break.
                blocks = tuple(get_best_3d_decomposition(total_domain_size))
                total_domain_size = prod(initial_cells_per_block) * processes_per_node
                cells_per_block = tuple(get_best_3d_decomposition(total_domain_size, descending=True))

    outer_iterations = max(1, int(0.85 * num_processes / PROCESSES_PER_NODE)) # C * #nodes
    # |-> C=0.85: `slowdown factor` to ensure that for many nodes, the run does not get too long.

    for t in ["simpleOverlap"]:
        for split in [(1, 1, 1), (8, 8, 8), (16, 1, 1)]:
            if all(2 * s < c for s, c in zip(split, cells_per_block)):
                scenarios.add(Scenario(benchmark_name="strongScalingUniformGrid",
                                       blocks=blocks,
                                       cells_per_block=tuple(cells_per_block),
                                       time_step_strategy=t,
                                       inner_outer_split=tuple(split),
                                       boundary_setup=True,
                                       outer_iterations=2*outer_iterations,
                                       timesteps=TIME_STEPS_TO_EXECUTE,
                ))
            else:
                wlb.log_warning_on_root(f"Split {split} would exceed cells per block {cells_per_block}. Skipping.")


def single_node_benchmark():
    """Benchmarks only the LBM compute kernel"""
    wlb.log_info_on_root("Running single node benchmarks")
    wlb.log_info_on_root("")

    scenarios = wlb.ScenarioManager()

    inner_outer_splits = [(1, 1, 1), (8, 8, 8), (16, 1, 1)]

    for block_size in DOMAIN_BLOCK_SIZES_SINGLE_NODE:
        for split in inner_outer_splits:
            if all(2 * s < c for s, c in zip(split, block_size)):
                if split == inner_outer_splits[0]:
                    time_step_strategies = ["simpleOverlap", "kernelOnly", "StreamOnly"]
                else:
                    time_step_strategies = ["simpleOverlap"]
                for t in time_step_strategies:
                    scenario = Scenario(benchmark_name="singleNodeUniformGrid",
                                        cells_per_block=tuple(block_size),
                                        time_step_strategy=t,
                                        inner_outer_split=tuple(split),
                                        boundary_setup=True,
                                        outer_iterations=2,
                                        timesteps=num_time_steps(block_size),
                                        )
                    scenarios.add(scenario)
            else:
                wlb.log_warning_on_root(f"Split {split} would exceed cells per block {block_size}. Skipping.")


def validation_run():
    """Run with full periodic shear flow or boundary scenario (ldc) to check if the code works"""
    wlb.log_info_on_root("Validation run")
    wlb.log_info_on_root("")

    time_step_strategy = "noOverlap"  # "noOverlap"

    scenarios = wlb.ScenarioManager()
    scenario = Scenario(benchmark_name="validationUniformGrid",
                        cells_per_block=(64, 64, 64),
                        time_step_strategy=time_step_strategy,
                        timesteps=201,
                        outer_iterations=1,
                        warmup_steps=0,
                        init_shear_flow=False,
                        boundary_setup=True,
                        vtk_write_frequency=5000,
                        remaining_time_logger_frequency=30)
    scenarios.add(scenario)


wlb.log_info_on_root(f"Batch run of benchmark scenarios, saving result to {DB_FILE}")
# Select the benchmark you want to run
wlb.log_info_on_root("Initializing benchmarks...")
if BENCHMARK == -1:
    validation_run()
else:
    if not set_parameter_for_specific_cluster():
        wlb.log_error_on_root("Cluster specific simulation environment could not be set. Exiting.")
        sys.exit(1)

    wlb.log_info_on_root("Environment parameters used for the benchmark:" +
                         f"\n  TIME_STEPS_FOR_128_BLOCK: {TIME_STEPS_FOR_128_BLOCK}" +
                         f"\n  PROCESSES_PER_NODE:       {PROCESSES_PER_NODE}" +
                         f"\n  TIME_STEPS_TO_EXECUTE:    {TIME_STEPS_TO_EXECUTE}" +
                         f"\n  MAIN_MEMORY_GB:           {MAIN_MEMORY_GB}" +
                         f"\n  NUM_CORES_PER_NODE:       {NUM_CORES_PER_NODE}" +
                         f"\n  MAX_AVAIL_NODES:          {MAX_AVAIL_NODES}" +
                         f"\n  WeakX:                    {WeakX}" +
                         f"\n  WeakY:                    {WeakY}" +
                         f"\n  WeakZ:                    {WeakZ}" +
                         f"\n  StrongDecompositionX:     {StrongDecompositionX}" +
                         f"\n  StrongDecompositionY:     {StrongDecompositionY}" +
                         f"\n  StrongDecompositionZ:     {StrongDecompositionZ}" +
                         f"\n  DOMAIN_BLOCK_SIZES_SINGLE_NODE: {DOMAIN_BLOCK_SIZES_SINGLE_NODE}")
    wlb.log_info_on_root("Cluster specific simulation environment successfully set.")

    if BENCHMARK == 0:
        # benchmarks different sizes for communication layer, and domain sizes on single CPU
        single_node_benchmark()
    elif BENCHMARK == 1:
        weak_scaling_benchmark()
    elif BENCHMARK == 2:
        strong_scaling_benchmark()
    else:
        wlb.log_error_on_root(f"Benchmark setup (BENCHMARK: {str(BENCHMARK)}) not defined. Benchmark execution skipped.")