import os
import sys
import sqlite3
from math import prod, log2
from numpy import argsort
import waLBerla as wlb
from waLBerla.tools.config import get_best_3d_decomposition
from waLBerla.tools.sqlitedb import sequenceValuesToScalars, checkAndUpdateSchema, storeSingle

try:
    import machinestate as ms
except ImportError:
    ms = None

# Define global variables with default values
BENCHMARK = int(os.environ.get('BENCHMARK', 0))
CLUSTER = str(os.environ.get('CLUSTER', None))
DB_FILE = str(os.environ.get('DB_FILE', "gpu_benchmark.sqlite3"))
MINIMAL_RUNTIME_SINGLE_NODE = int(os.environ.get('MINIMAL_RUNTIME_SINGLE_NODE', 5))

TIME_STEPS_FOR_128_BLOCK = None
TIME_STEPS_TO_EXECUTE = None
PROCESSES_PER_NODE = None
MAIN_MEMORY_GB, NUM_CORES_PER_NODE, MAX_AVAIL_NODES = None, None, None
GPU_MEMORY_GB, NUM_GPUS_PER_NODE = None, None
MPI_AWARE_COMMUNICATION = None
WeakX, WeakY, WeakZ = None, None, None
StrongDecompositionX, StrongDecompositionY, StrongDecompositionZ = None, None, None
DOMAIN_BLOCK_SIZES_SINGLE_NODE = None


def set_parameter_for_specific_cluster():
    global TIME_STEPS_FOR_128_BLOCK, TIME_STEPS_TO_EXECUTE, PROCESSES_PER_NODE
    global MAIN_MEMORY_GB, NUM_CORES_PER_NODE, MAX_AVAIL_NODES
    global GPU_MEMORY_GB, NUM_GPUS_PER_NODE
    global MPI_AWARE_COMMUNICATION
    global WeakX, WeakY, WeakZ
    global StrongDecompositionX, StrongDecompositionY, StrongDecompositionZ
    global DOMAIN_BLOCK_SIZES_SINGLE_NODE

    cluster_mapping = {
        "marenostrum5": "MareNostrum5",
        "marenostrum": "MareNostrum5",
        "lumi": "LUMI"
    }

    normalized_cluster = (CLUSTER or "MareNostrum5").lower() # MareNostrum5 is the default choice if CLUSTER=None
    matched_cluster = cluster_mapping.get(normalized_cluster)

    if matched_cluster == "MareNostrum5":
        TIME_STEPS_FOR_128_BLOCK = int(os.environ.get('TIME_STEPS_FOR_128_BLOCK', 540))
        PROCESSES_PER_NODE = int(os.environ.get('PROCESSES_PER_NODE', 4))
        TIME_STEPS_TO_EXECUTE = int(os.environ.get('TIME_STEPS_TO_EXECUTE', 20))
        MAIN_MEMORY_GB = int(os.environ.get('MAIN_MEMORY_GB', 256)) # Default for MareNostrum5-GPP
        GPU_MEMORY_GB = int(os.environ.get('GPU_MEMORY_GB', 64)) # Default for MareNostrum5-ACC
        NUM_CORES_PER_NODE = int(os.environ.get('NUM_CORES_PER_NODE', 112)) # Default for MareNostrum5-GPP
        NUM_GPUS_PER_NODE = int(os.environ.get('NUM_GPUS_PER_NODE', 4)) # Default for MareNostrum5-ACC
        MAX_AVAIL_NODES = int(os.environ.get('MAX_AVAIL_NODES', 100)) # Default for MareNostrum5-GPP
        MPI_AWARE_COMMUNICATION = bool(os.environ.get('MPI_AWARE_COMMUNICATION', True)) # Default for MareNostrum5-ACC

        WeakX = int(os.environ.get('WeakX', 640))
        WeakY = int(os.environ.get('WeakY', 640))
        WeakZ = int(os.environ.get('WeakZ', 640))

        StrongDecompositionX = int(os.environ.get('StrongX', 1))
        StrongDecompositionY = int(os.environ.get('StrongY', 2))
        StrongDecompositionZ = int(os.environ.get('StrongZ', 2))

        DOMAIN_BLOCK_SIZES_SINGLE_NODE = [
            (704, 704, 704),    # N=3.489e+08 ~ 99.4% |  1.0%
            (640, 640, 640),    # N=2.621e+08 ~ 74.7% |  1.0% # 100 nodes included
            (576, 576, 576),    # N=1.911e+08 ~ 54.5% |  1.0%
            (512, 512, 512),    # N=1.342e+08 ~ 38.3% |  1.0%
            (384, 384, 384),    # N=5.662e+07 ~ 16.1% |  1.0%
            (320, 320, 320),    # N=3.277e+07 ~  9.3% |  1.0% # 100 nodes included
        ]
        return True

    elif matched_cluster == "LUMI":
        TIME_STEPS_FOR_128_BLOCK = int(os.environ.get('TIME_STEPS_FOR_128_BLOCK', 442))
        PROCESSES_PER_NODE = int(os.environ.get('PROCESSES_PER_NODE', 8))
        TIME_STEPS_TO_EXECUTE = int(os.environ.get('TIME_STEPS_TO_EXECUTE', 20))
        MAIN_MEMORY_GB = int(os.environ.get('MAIN_MEMORY_GB', 256)) # Default for Lumi-C
        GPU_MEMORY_GB = int(os.environ.get('GPU_MEMORY_GB', 64)) # Default for Lumi-G
        NUM_CORES_PER_NODE = int(os.environ.get('NUM_CORES_PER_NODE', 128)) # Default for Lumi-C
        NUM_GPUS_PER_NODE = int(os.environ.get('NUM_GPUS_PER_NODE', 8)) # Default for Lumi-G
        MAX_AVAIL_NODES = int(os.environ.get('MAX_AVAIL_NODES', 512)) # Default for Lumi-C
        MPI_AWARE_COMMUNICATION = bool(os.environ.get('MPI_AWARE_COMMUNICATION', True)) # Default for Lumi-G

        WeakX = int(os.environ.get('WeakX', 640))
        WeakY = int(os.environ.get('WeakY', 640))
        WeakZ = int(os.environ.get('WeakZ', 640))

        StrongDecompositionX = int(os.environ.get('StrongX', 2))
        StrongDecompositionY = int(os.environ.get('StrongY', 2))
        StrongDecompositionZ = int(os.environ.get('StrongZ', 2))

        DOMAIN_BLOCK_SIZES_SINGLE_NODE = [
            (704, 704, 704),    # N=6.978e+08 ~ 99.4% |  1.0%
            (640, 640, 640),    # N=5.243e+08 ~ 74.7% |  1.0%
            (576, 576, 576),    # N=3.822e+08 ~ 54.5% |  1.0%
            (512, 512, 512),    # N=2.684e+08 ~ 38.3% |  1.0%
            (384, 384, 384),    # N=1.132e+08 ~ 16.1% |  1.0%
            (256, 256, 256),    # N=3.355e+07 ~  4.8% |  1.0%
            (128, 128, 128),    # N=4.194e+06 ~  0.6% |  1.0%
        ]

        return True

    else:
        wlb.log_warning_on_root(f"Cluster '{str(CLUSTER)}' is not recognized. Benchmark execution skipped.")
        return False



def num_time_steps(block_size, time_steps_for_128_block=None):
    """
    Calculate the number of time steps based on the block size.

    This function computes the number of time steps required for a given block size
    by scaling the time steps that could be executed on one process within one second
    for a 128x128x128 cells_per_block to the given cells_per_block size.

    Parameters:
    block_size (tuple): A tuple of three integers representing the dimensions of the cells_per_block (x, y, z).
    time_steps_for_128_block (int, optional): The number of time steps for a 128x128x128 block.

    Returns:
    int: The calculated number of time steps, with a minimum value of 5.
    """
    if time_steps_for_128_block is None:
        time_steps_for_128_block = TIME_STEPS_FOR_128_BLOCK
    if time_steps_for_128_block is None:
        raise ValueError("TIME_STEPS_FOR_128_BLOCK is not initialized.")

    cells = block_size[0] * block_size[1] * block_size[2]
    time_steps = (128 ** 3 / cells) * time_steps_for_128_block * MINIMAL_RUNTIME_SINGLE_NODE
    if time_steps < 5:
        time_steps = 5
    return int(time_steps)


ldc_setup = {'Border': [
    {'direction': 'N', 'walldistance': -1, 'flag': 'UBB'},
    {'direction': 'W', 'walldistance': -1, 'flag': 'NoSlip'},
    {'direction': 'E', 'walldistance': -1, 'flag': 'NoSlip'},
    {'direction': 'S', 'walldistance': -1, 'flag': 'NoSlip'},
    {'direction': 'B', 'walldistance': -1, 'flag': 'NoSlip'},
    {'direction': 'T', 'walldistance': -1, 'flag': 'NoSlip'},
]}


def cuda_block_size_ok(block_size, regs_per_threads=168):
    """
    Checks if a given CUDA block size does not exceed the SM register limit.
    168 registers per thread was obtained using `cuobjdump` on both SRT and Cumulant
    kernels. You might want to validate that for your own kernels.
    """
    return prod(block_size) * regs_per_threads < 64 * (2 ** 10) # The H100 has 64K 32-bit registers


def domain_block_size_ok(block_size, total_mem, gls=1, q=27, size_per_value=8):
    """
    Checks if a single block of given size fits into GPU memory.
    """
    return prod(b + 2 * gls for b in block_size) * q * size_per_value < total_mem


class Scenario:
    def __init__(self, benchmark_name="not defined", blocks=None, cells_per_block=(128, 128, 128), periodic=(1, 1, 1), cuda_blocks=(128, 1, 1),
                 timesteps=None, time_step_strategy="normal", omega=1.8, mpi_aware_communication=False,
                 inner_outer_split=(1, 1, 1), warmup_steps=5, outer_iterations=3,
                 init_shear_flow=False, boundary_setup=False,
                 vtk_write_frequency=0, remaining_time_logger_frequency=-1,
                 additional_info=None, db_file_name=None):

        self.benchmark_name = benchmark_name

        if boundary_setup:
            init_shear_flow = False
            periodic = (0, 0, 0)

        self.blocks = tuple(blocks) if blocks else tuple(get_best_3d_decomposition(wlb.mpi.numProcesses()))

        self.cells_per_block = tuple(cells_per_block)
        self.periodic = tuple(periodic)

        self.time_step_strategy = time_step_strategy
        self.omega = omega
        self.timesteps = timesteps if timesteps else num_time_steps(cells_per_block)
        self.mpi_aware_communication = mpi_aware_communication
        self.inner_outer_split = tuple(inner_outer_split)
        self.init_shear_flow = init_shear_flow
        self.boundary_setup = boundary_setup
        self.warmup_steps = warmup_steps
        self.outer_iterations = outer_iterations
        self.cuda_blocks = tuple(cuda_blocks)

        self.vtk_write_frequency = vtk_write_frequency
        self.remaining_time_logger_frequency = remaining_time_logger_frequency
        self.db_file_name = DB_FILE if db_file_name is None else db_file_name

        self.config_dict = self.config(print_dict=False)
        self.additional_info = additional_info

    @wlb.member_callback
    def config(self, print_dict=True):
        from pprint import pformat
        config_dict = {
            'DomainSetup': {
                'blocks': self.blocks,
                'cellsPerBlock': self.cells_per_block,
                'periodic': self.periodic,
            },
            'Parameters': {
                'omega': self.omega,
                'mpiAwareCommunication': self.mpi_aware_communication,
                'warmupSteps': self.warmup_steps,
                'outerIterations': self.outer_iterations,
                'timeStepStrategy': self.time_step_strategy,
                'timesteps': self.timesteps,
                'initShearFlow': self.init_shear_flow,
                'gpuBlockSize': self.cuda_blocks,
                'innerOuterSplit': self.inner_outer_split,
                'vtkWriteFrequency': self.vtk_write_frequency,
                'remainingTimeLoggerFrequency': self.remaining_time_logger_frequency
            },
            'Logging': {
                'logLevel': 'info',  # info progress detail tracing
            }
        }
        if self.boundary_setup:
            config_dict["Boundaries"] = ldc_setup

        if print_dict:
            wlb.log_info_on_root("\n" + 50*"=" + "\n")
            wlb.log_info_on_root("Scenario:\n" + pformat(config_dict))
            if self.additional_info:
                wlb.log_info_on_root("Additional Info:\n" + pformat(self.additional_info))
        return config_dict

    @wlb.member_callback
    def results_callback(self, **kwargs):
        data = {}
        data.update(self.config_dict['Parameters'])
        data.update(self.config_dict['DomainSetup'])
        data.update(kwargs)

        if self.additional_info is not None:
            data.update(self.additional_info)

        data['Identifier'] = self.benchmark_name
        data['executable'] = sys.argv[0]
        data['compile_flags'] = wlb.build_info.compiler_flags
        data['walberla_version'] = wlb.build_info.version
        data['build_machine'] = wlb.build_info.build_machine

        if ms:
            state = ms.MachineState(extended=False, anonymous=True)
            state.generate()  # generate subclasses
            state.update()  # read information
            data["MachineState"] = str(state.get())
        else:
            print("MachineState module is not available. MachineState was not saved")

        sequenceValuesToScalars(data)

        result = data
        sequenceValuesToScalars(result)
        num_tries = 4
        table_name = "timeStepping_" + str(self.time_step_strategy)
        # table_name = f"GPU_runs_{data['stencil']}_{data['streamingPattern']}_{data['collisionSetup']}_{prod(self.blocks)}nodes_{'x'.join(map(str, self.cells_per_block))}cellsPerBlock"
        table_name = table_name.replace("-", "_")
        table_name = table_name.replace(" ", "_")

        for num_try in range(num_tries):
            try:
                checkAndUpdateSchema(result, table_name, self.db_file_name, alter_table=True)
                storeSingle(result, table_name, self.db_file_name)
                break
            except sqlite3.OperationalError as e:
                wlb.log_warning(f"Sqlite DB writing failed: try {num_try + 1}/{num_tries}  {str(e)}")


# -------------------------------------- Functions trying different parameter sets -----------------------------------


def weak_scaling_benchmark():
    wlb.log_info_on_root("Running scaling benchmark with one block per proc and communication hiding")
    wlb.log_info_on_root("")

    scenarios = wlb.ScenarioManager()
    cells_per_block = (WeakX, WeakY, WeakZ)

    for t in ["simpleOverlap"]:
        for split in [(1, 1, 1), (8, 8, 8), (16, 1, 1)]:
            if all(2 * s < c for s, c in zip(split, cells_per_block)):
                scenarios.add(Scenario(benchmark_name="weakScalingUniformGrid",
                                       cells_per_block=tuple(cells_per_block),
                                       cuda_blocks=(128, 1, 1),
                                       time_step_strategy=t,
                                       inner_outer_split=tuple(split),
                                       mpi_aware_communication=MPI_AWARE_COMMUNICATION,
                                       boundary_setup=True,
                                       outer_iterations=2,
                                       timesteps=TIME_STEPS_TO_EXECUTE,
                ))
            else:
                wlb.log_warning_on_root(f"Split {split} would exceed cells per block {cells_per_block}. Skipping.")


def strong_scaling_benchmark():
    wlb.log_info_on_root("Running strong scaling benchmark with one block per proc with communication hiding")
    wlb.log_info_on_root("")

    scenarios = wlb.ScenarioManager()

    num_processes = wlb.mpi.numProcesses()
    initial_blocks = [StrongDecompositionX, StrongDecompositionY, StrongDecompositionZ]
    processes_per_node = prod(initial_blocks)
    assert num_processes % processes_per_node == 0, "Number of processes must be a multiple of processes per node."
    number_of_nodes = num_processes // processes_per_node
    initial_cells_per_block = [WeakX, WeakY, WeakZ]
    scaling_step = log2(number_of_nodes)

    if not scaling_step.is_integer():
        blocks = None
        total_domain_size = prod(initial_cells_per_block) * processes_per_node
        cells_per_block = tuple(get_best_3d_decomposition(total_domain_size, descending=True))
    else: # This scaling step has a power of two number of nodes.
        blocks = initial_blocks
        cells_per_block = initial_cells_per_block
        for _ in range(int(scaling_step)):
            division_indices = argsort(cells_per_block)[::-1] # Sort in descending order
            for idx in division_indices:
                if cells_per_block[idx] % 2 == 0:
                    cells_per_block[idx] //= 2
                    blocks[idx] *= 2
                    break
            else: # else block runs only if the loop completes without encountering a break.
                blocks = tuple(get_best_3d_decomposition(total_domain_size))
                total_domain_size = prod(initial_cells_per_block) * processes_per_node
                cells_per_block = tuple(get_best_3d_decomposition(total_domain_size, descending=True))

    outer_iterations = max(1, int(0.85 * num_processes / NUM_GPUS_PER_NODE)) # C * #nodes
    # |-> C=0.85: `slowdown factor` to ensure that for many nodes, the run does not get too long.

    for t in ["simpleOverlap"]:
        for split in [(1, 1, 1), (8, 8, 8), (16, 1, 1)]:
            if all(2 * s < c for s, c in zip(split, cells_per_block)):
                scenarios.add(Scenario(benchmark_name="strongScalingUniformGrid",
                                       blocks=blocks,
                                       cells_per_block=tuple(cells_per_block),
                                       cuda_blocks=(128, 1, 1),
                                       time_step_strategy=t,
                                       inner_outer_split=tuple(split),
                                       mpi_aware_communication=MPI_AWARE_COMMUNICATION,
                                       boundary_setup=True,
                                       outer_iterations=2*outer_iterations,
                                       timesteps=TIME_STEPS_TO_EXECUTE,
                ))
            else:
                wlb.log_info_on_root(f"Split {split} would exceed cells per block {cells_per_block}. Skipping.")


def single_node_benchmark():
    """Benchmarks only the LBM compute kernel"""
    wlb.log_info_on_root("Running single node benchmarks")
    wlb.log_info_on_root("")

    # FIXME this is outdated. I set the device memory in the function in the beginning which would be totally be disregarded here.
    try:
        gpu_mem_gb = GPU_MEMORY_GB #int(os.environ.get('GPU_MEMORY_GB')) # QUICKFIX
        gpu_mem = gpu_mem_gb * (2 ** 30) # GPU Main Memory * the factor for GiB
    except TypeError:
        try:
            gpu_mem_mb = int(os.environ.get('GPU_MEMORY_MB'))
            gpu_mem = gpu_mem_mb * (2 ** 20) # GPU Main Memory * the factor for MiB
        except TypeError:
            try:
                gpu_mem_kb = int(os.environ.get('GPU_MEMORY_KB'))
                gpu_mem = gpu_mem_kb * (2 ** 10) # GPU Main Memory * the factor for KiB
            except TypeError:
                wlb.log_info_on_root("The size of device memory is unknown, so runtime errors can occur do to GPU out of memory.")
                gpu_mem = GPU_MEMORY_GB * (2 ** 30) # GPU Main Memory * the factor for GiB
    gpu_type = os.environ.get('GPU_TYPE')

    additional_info = {}
    if gpu_type is not None:
        additional_info['gpu_type'] = gpu_type

    scenarios = wlb.ScenarioManager()

    cuda_blocks = [(128, 1, 1), (32, 4, 1), (16, 4, 2), (8, 4, 4)]
    inner_outer_splits = [(1, 1, 1), (8, 8, 8), (16, 1, 1)]


    for block_size in DOMAIN_BLOCK_SIZES_SINGLE_NODE:
        for cuda_block_size in cuda_blocks:
            for split in inner_outer_splits:
                # cuda_block_size = (256, 1, 1) and block_size = (64, 64, 64) would be cut to cuda_block_size = (64, 1, 1)
                if cuda_block_size > block_size:
                    continue
                if not cuda_block_size_ok(cuda_block_size):
                    wlb.log_info_on_root(f"Cuda block size {cuda_block_size} would exceed register limit. Skipping.")
                    continue
                if not domain_block_size_ok(block_size, gpu_mem):
                    wlb.log_info_on_root(f"Block size {block_size} would exceed GPU memory. Skipping.")
                    continue
                if all(2 * s < c for s, c in zip(split, block_size)):
                    if split == inner_outer_splits[0] and cuda_block_size == cuda_blocks[0]:
                        time_step_strategies = ["simpleOverlap", "kernelOnly", "StreamOnly"]
                    else:
                        time_step_strategies = ["simpleOverlap"]
                    for t in time_step_strategies:
                        scenario = Scenario(benchmark_name="singleNodeUniformGrid",
                                            cells_per_block=tuple(block_size),
                                            cuda_blocks=tuple(cuda_block_size),
                                            time_step_strategy=t,
                                            inner_outer_split=tuple(split),
                                            additional_info=additional_info,
                                            mpi_aware_communication=MPI_AWARE_COMMUNICATION,
                                            boundary_setup=True,
                                            outer_iterations=2,
                                            timesteps=num_time_steps(block_size),
                                            )
                        scenarios.add(scenario)
                else:
                    wlb.log_warning_on_root(f"Split {split} would exceed cells per block {block_size}. Skipping.")


# TODO add a single GPU benchmark!


def validation_run():
    """Run with full periodic shear flow or boundary scenario (ldc) to check if the code works"""
    wlb.log_info_on_root("Validation run")
    wlb.log_info_on_root("")

    time_step_strategy = "noOverlap"  # "noOverlap"

    scenarios = wlb.ScenarioManager()
    scenario = Scenario(benchmark_name="validationUniformGrid",
                        cells_per_block=(128, 128, 128),
                        time_step_strategy=time_step_strategy,
                        timesteps=201,
                        outer_iterations=1,
                        warmup_steps=0,
                        init_shear_flow=False,
                        boundary_setup=True,
                        vtk_write_frequency=5000,
                        remaining_time_logger_frequency=30)
    scenarios.add(scenario)


wlb.log_info_on_root(f"Batch run of benchmark scenarios, saving result to {DB_FILE}")
# Select the benchmark you want to run
wlb.log_info_on_root("Initializing benchmarks...")
if BENCHMARK == -1:
    validation_run()
else:
    if not set_parameter_for_specific_cluster():
        wlb.log_error_on_root("Cluster specific simulation environment could not be set. Exiting.")
        sys.exit(1)

    wlb.log_info_on_root("Environment parameters used for the benchmark:" +
                         f"\n  TIME_STEPS_FOR_128_BLOCK: {TIME_STEPS_FOR_128_BLOCK}" +
                         f"\n  PROCESSES_PER_NODE:       {PROCESSES_PER_NODE}" +
                         f"\n  TIME_STEPS_TO_EXECUTE:    {TIME_STEPS_TO_EXECUTE}" +
                         f"\n  MAIN_MEMORY_GB:           {MAIN_MEMORY_GB}" +
                         f"\n  NUM_CORES_PER_NODE:       {NUM_CORES_PER_NODE}" +
                         f"\n  MAX_AVAIL_NODES:          {MAX_AVAIL_NODES}" +
                         f"\n  GPU_MEMORY_GB:            {GPU_MEMORY_GB }" +
                         f"\n  NUM_GPUS_PER_NODE:        {NUM_GPUS_PER_NODE}" +
                         f"\n  MPI_AWARE_COMMUNICATION:  {MPI_AWARE_COMMUNICATION}" +
                         f"\n  WeakX:                    {WeakX}" +
                         f"\n  WeakY:                    {WeakY}" +
                         f"\n  WeakZ:                    {WeakZ}" +
                         f"\n  StrongDecompositionX:     {StrongDecompositionX}" +
                         f"\n  StrongDecompositionY:     {StrongDecompositionY}" +
                         f"\n  StrongDecompositionZ:     {StrongDecompositionZ}" +
                         f"\n  DOMAIN_BLOCK_SIZES_SINGLE_NODE: {DOMAIN_BLOCK_SIZES_SINGLE_NODE}")
    wlb.log_info_on_root("Cluster specific simulation environment successfully set.")

    if BENCHMARK == 0:
        # benchmarks different CUDA block sizes, sizes for communication layer, and domain sizes on single GPU
        single_node_benchmark()
    elif BENCHMARK == 1:
        weak_scaling_benchmark()
    elif BENCHMARK == 2:
        strong_scaling_benchmark()
    else:
        wlb.log_error_on_root(f"Benchmark setup (BENCHMARK: {str(BENCHMARK)}) not defined. Benchmark execution skipped.")